@@ -1,26 +1,6 @@
 # -*- encoding: utf-8 -*-
-###############################################################################
-#
-# Copyright (c) 2014-2017 Cedric Krier.
-# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
-# Reserved.
-#
-# This program is free software; you can redistribute it and/or modify it under
-# the terms of the GNU General Public License as published by the Free Software
-# Foundation; either version 3 of the License, or (at your option) any later
-# version.
-#
-# This program is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
-# details.
-#
-# You should have received a copy of the GNU General Public License along with
-# this program.  If not, see <http://www.gnu.org/licenses/>.
-#
-###############################################################################
-
-
+# This file is part of relatorio.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import os
 import unittest
 from io import StringIO, BytesIO
@@ -42,16 +22,16 @@ def pseudo_gettext(string):
                'Bonjour,': 'Hello,',
                'Je suis un test de templating en odt.':
                'I am an odt templating test',
-               'Felix da housecat': u'Félix le chat de la maison',
-               'We sell stuff': u'On vend des choses',
+               'Felix da housecat': 'Félix le chat de la maison',
+               'We sell stuff': 'On vend des choses',
               }
     return catalog.get(string, string)
 
 
 def stream_to_string(stream):
     # In Python 3, stream will be bytes
-    if not isinstance(stream, unicode):
-        return unicode(stream, 'utf-8')
+    if not isinstance(stream, str):
+        return str(stream, 'utf-8')
     return stream
 
 
@@ -62,26 +42,31 @@ class TestOOTemplating(unittest.TestCase):
         filepath = os.path.join(thisdir, 'test.odt')
         self._source = open(filepath, mode='rb')
         self.oot = Template(self._source)
-        self.data = {'first_name': u'Trente',
-                     'last_name': u'Møller',
-                     'ville': u'Liège',
-                     'friends': [{'first_name': u'Camille',
-                                  'last_name': u'Salauhpe'},
-                                 {'first_name': u'Mathias',
-                                  'last_name': u'Lechat'}],
-                     'hobbies': [u'Music', u'Dancing', u'DJing'],
-                     'animals': [u'Felix da housecat', u'Dog eat Dog'],
+        self.data = {'first_name': 'Trente',
+                     'last_name': 'Møller',
+                     'ville': 'Liège',
+                     'friends': [{'first_name': 'Camille',
+                                  'last_name': 'Salauhpe'},
+                                 {'first_name': 'Mathias',
+                                  'last_name': 'Lechat'}],
+                     'hobbies': ['Music', 'Dancing', 'DJing'],
+                     'animals': ['Felix da housecat', 'Dog eat Dog'],
                      'images': [(open(os.path.join(thisdir, 'one.jpg'), 'rb'),
                                  'image/jpeg', None, None, 'One'),
                                 (open(os.path.join(thisdir, 'two.png'), 'rb'),
-                                 'image/png', '2cm', '2.2cm', 'Two')],
+                                 'image/png', '2cm', '2.2cm', 'Two'),
+                                (None, None)],
                      'oeuf': open(os.path.join(thisdir, 'egg.jpg'), 'rb'),
-                     'footer': u'We sell stuff'}
+                     'footer': 'We sell stuff',
+                     'salutation': 'Greating',
+                     'title': 'Testing Letter',
+                     }
 
     def tearDown(self):
         self._source.close()
         for image in self.data['images']:
-            image[0].close()
+            if image[0]:
+                image[0].close()
         self.data['oeuf'].close()
 
     def test_init(self):
@@ -121,11 +106,12 @@ class TestOOTemplating(unittest.TestCase):
         cell = root_interpolated[0]
         self.assertRegex(
             cell.get('{http://genshi.edgewall.org/}attrs'),
-            '__relatorio_guess_type\(__relatorio_store_cache\([0-9]*, foo\)\)')
+            r'__relatorio_guess_type\(__relatorio_store_cache\([0-9]*, foo\)\)'
+            )
         child = cell[0][0]
         self.assertRegex(
             child.get('{http://genshi.edgewall.org/}replace'),
-            '__relatorio_get_cache\([0-9]*\)')
+            r'__relatorio_get_cache\([0-9]*\)')
 
     def test_directives_in_cell_with_children(self):
         "Testing directive inside cell with children"
@@ -292,7 +278,7 @@ class TestOOTemplating(unittest.TestCase):
         <table:covered-table-cell/>
         <table:covered-table-cell/>
     </table:table-row>
-</table:table>'''
+</table:table>'''  # noqa: E501
         interpolated = self.oot.insert_directives(xml)
         root = lxml.etree.parse(interpolated).getroot()
         child2 = root[1]
@@ -336,7 +322,7 @@ class TestOOTemplating(unittest.TestCase):
         "Test warning for missing statement text"
         xml = b'''<xml xmlns:text="urn:text" xmlns:xlink="urn:xlink">
                     <text:a xlink:href="relatorio://content:foo">content:bar</text:a>
-                  </xml>'''
+                  </xml>'''  # noqa: E501
 
         with self.assertWarnsRegex(
                 UserWarning,
@@ -352,7 +338,7 @@ class TestOOTemplating(unittest.TestCase):
         xml = b'''<xml xmlns:text="urn:text" xmlns:xlink="urn:xlink">
                     <text:a xlink:href="relatorio://with foo='test'">with foo='test'</text:a>
                     <text:a xlink:href="relatorio:///with"></text:a>
-                  </xml>'''
+                  </xml>'''  # noqa: E501
 
         with self.assertWarnsRegex(
                 UserWarning,
@@ -368,7 +354,7 @@ class TestOOTemplating(unittest.TestCase):
                     zzz
                     <text:a xlink:href="relatorio:///if">/if</text:a>
                     aaa
-                 </xml>'''
+                 </xml>'''  # noqa: E501
         interpolated = self.oot.insert_directives(xml)
         root_interpolated = lxml.etree.parse(interpolated).getroot()
         child = root_interpolated[0]
@@ -388,13 +374,20 @@ class TestOOTemplating(unittest.TestCase):
         self.assertRaises(UndefinedError,
             lambda: stream.events.render(encoding='utf-8'))
 
+    def test_meta(self):
+        "Testing that meta get rendered"
+        stream = self.oot.generate(**self.data)
+        rendered = stream_to_string(stream.events.render(encoding='utf-8'))
+        self.assertTrue("Greating" in rendered)
+        self.assertTrue("Testing Letter" in rendered)
+
     def test_generate(self):
         "Testing that content get rendered"
         stream = self.oot.generate(**self.data)
         rendered = stream_to_string(stream.events.render(encoding='utf-8'))
         self.assertTrue('Bonjour,' in rendered)
         self.assertTrue('Trente' in rendered)
-        self.assertTrue(u'Møller' in rendered)
+        self.assertTrue('Møller' in rendered)
         self.assertTrue('Dog eat Dog' in rendered)
         self.assertTrue('Felix da housecat' in rendered)
 
@@ -407,7 +400,7 @@ class TestOOTemplating(unittest.TestCase):
         self.assertTrue("Hello," in translated_xml)
         self.assertTrue("I am an odt templating test" in translated_xml)
         self.assertTrue('Felix da housecat' not in translated_xml)
-        self.assertTrue(u'Félix le chat de la maison' in translated_xml)
+        self.assertTrue('Félix le chat de la maison' in translated_xml)
         self.assertTrue('We sell stuff' not in translated_xml)
         self.assertTrue('On vend des choses' in translated_xml)
 
@@ -419,22 +412,24 @@ class TestOOTemplating(unittest.TestCase):
         tree = lxml.etree.parse(StringIO(rendered[25:styles_idx]))
         root = tree.getroot()
         images = root.xpath('//draw:frame', namespaces=self.oot.namespaces)
-        self.assertEqual(len(images), 3)
+        self.assertEqual(len(images), 5)
         self.assertFalse(
             images[0].get('{%s}name' % self.oot.namespaces['draw']))
+        self.assertFalse(
+            images[1].get('{%s}name' % self.oot.namespaces['draw']))
         self.assertEqual(
-            images[1].get('{%s}name' % self.oot.namespaces['draw']), 'One')
+            images[2].get('{%s}name' % self.oot.namespaces['draw']), 'One')
         self.assertEqual(
-            images[2].get('{%s}name' % self.oot.namespaces['draw']), 'Two')
+            images[3].get('{%s}name' % self.oot.namespaces['draw']), 'Two')
         self.assertEqual(
-            images[1].get('{%s}width' % self.oot.namespaces['svg']), '1.732cm')
-        self.assertEqual(
-            images[1].get('{%s}height' % self.oot.namespaces['svg']),
-            '1.513cm')
-        self.assertEqual(
-            images[2].get('{%s}width' % self.oot.namespaces['svg']), '2cm')
+            images[2].get('{%s}width' % self.oot.namespaces['svg']), '1.732cm')
         self.assertEqual(
             images[2].get('{%s}height' % self.oot.namespaces['svg']),
+            '1.513cm')
+        self.assertEqual(
+            images[3].get('{%s}width' % self.oot.namespaces['svg']), '2cm')
+        self.assertEqual(
+            images[3].get('{%s}height' % self.oot.namespaces['svg']),
             '2.2cm')
 
     def test_regexp(self):
